#include <stdio.h>
#include <jpeglib.h>
#include <assert.h>
#include "libjpegrest.h"

/* Prototypes*/
static int
bjpeg_read(struct jpeg_decompress_struct * srcinfo,
		struct bjpeginfo *jpeg);


struct bjpeginfo
bjpeg_alloc(const int nplanes, const int *widths, const int *heights)
{
	struct bjpeginfo jpeg;
	int bplane, brow;
	int memsize;
	char * offset;

	/* We should allocate all the size for the arrays at once. That is:
	 * sizeof(cblockplane) * nplanes
	 * + sizeof(cblockrow) * brows[bplane]
	 * + sizeof(cblock) * bcolumns[bplane][brows]
	 */
	memsize=0;
	memsize+=sizeof(cblockplane) * nplanes;
	for (bplane = 0; bplane < nplanes; ++bplane)
	{
		memsize+=sizeof(cblockrow) * heights[bplane];
		memsize+=sizeof(cblock) * widths[bplane] * heights[bplane];
	}
	printf("Allocating for %i bytes.\n", memsize);

	jpeg.blocks = (cblockplane *) malloc(memsize);
	assert(jpeg.blocks != NULL);

	/* Fill the structure */
	jpeg.nplanes = nplanes;
	offset = (char *) jpeg.blocks;
	/* Reserve for the cblockplane pointers */
	offset += sizeof(cblockplane) * nplanes;
	/* Reserve for the cblockplane rows */
	for (bplane = 0; bplane < nplanes; ++bplane)
	{
		jpeg.blocks[bplane] = (cblockplane) offset;
		offset += sizeof(cblockrow) * heights[bplane];
		for(brow = 0; brow < heights[bplane]; ++brow)
		{
			jpeg.blocks[bplane][brow] = (cblockrow) offset;
			offset+=sizeof(cblock) * widths[bplane];
		}
	}
	printf("End offset: %i bytes.\n", offset - (char *)jpeg.blocks);


	/* Allocate for qtables */
	/* ********** TO BE DONE ************/
	jpeg.qtables = NULL;

	return jpeg;
}

void 
bjpeg_free(struct bjpeginfo * jpeg)
{
	if(jpeg->blocks != NULL)
		free(jpeg->blocks);
	if(jpeg->qtables != NULL)
		free(jpeg->qtables);
}

void bjpeg_init(int *argn, char **argv)
{
	return;
}

void bjpeg_close(struct bjpeginfo *jpeg)
{
	jpeg_destroy_decompress(&jpeg->srcinfo);
	fclose(jpeg->io_file);
}

int
bjpeg_filename_get(const char *filename, struct bjpeginfo *jpeg)
{
	FILE *input;
	int res;

	/* Opening the FILE */
	input = fopen(filename,"r");
	if (input == NULL)
	{
		fprintf(stderr,"The file \"%s\" cannot be opened.\n",
			filename);
		res = -1;
	} else
	{
		res = bjpeg_stdio_get(input, jpeg);
	}
	
	return res;
}

int
bjpeg_stdio_get(FILE *input, struct bjpeginfo *jpeg)
{
	struct jpeg_decompress_struct srcinfo;
	struct jpeg_error_mgr jsrcerr;

	srcinfo.err = jpeg_std_error(&jsrcerr);

	jpeg_create_decompress(&srcinfo);

	jpeg_stdio_src(&srcinfo, input);
	jpeg_read_header(&srcinfo, TRUE);

	bjpeg_read(&srcinfo, jpeg);
	jpeg->io_file = input;
	
	return 0;
}

static int
bjpeg_read(struct jpeg_decompress_struct * srcinfo,
		struct bjpeginfo *jpeg)
{
	int brow, bcolumn, bplane, coef;
	int widths_in_blocks[4], heights_in_blocks[4]; /* 4 max. for each
		component. */
	/* jvirt_barray_ptr => struct jvirt_barray_control
	 *
	 * There's one jvirt_barray_ptr for each component, accessed
	 * through coef_ptr[component]. A function in the JPEG library
	 * can return a JBLOCKARRAY for each jvirt_barray_ptr.*/
	jvirt_barray_ptr * vblock_ptr; /* JPEG coefficients blocks varray.*/
	/* typedef JCOEF JBLOCK[DCTSIZE2];
	 * typedef JBLOCK FAR * JBLOCKROW;
	 * typedef JBLOCKROW * JBLOCKARRAY; */
	JBLOCKARRAY block_ptr;  /* All the blocks of a plane */

	assert(srcinfo->out_color_components <= 4); /* 4, as declared. */
	/* Set up the component widths/heights in an array */
	for (bplane = 0; bplane < srcinfo->num_components; ++bplane)
	{
		widths_in_blocks[bplane] =
			srcinfo->comp_info[bplane].width_in_blocks;
		heights_in_blocks[bplane] =
			srcinfo->comp_info[bplane].height_in_blocks;
	}
	*jpeg = bjpeg_alloc(srcinfo->num_components,
			widths_in_blocks,
			heights_in_blocks);

	/* Get memory for the buffer scanline pointers */
	vblock_ptr = jpeg_read_coefficients(srcinfo);
	for (bplane = 0; bplane < srcinfo->num_components; ++bplane)
	{

		for(brow = 0; brow < heights_in_blocks[bplane]; ++brow)
		{
			block_ptr = (*srcinfo->mem->access_virt_barray)
				((j_common_ptr) srcinfo, vblock_ptr[bplane],
				brow, 1, 0); /* Last is FALSE */
			for(bcolumn = 0; bcolumn < widths_in_blocks[bplane];
				++bcolumn)
				for(coef = 0; coef < 64; coef++)
				{
					jpeg->blocks[bplane][brow][bcolumn]
						[coef] =
						block_ptr[0][bcolumn][coef];
				}
		}
	}

	jpeg->srcinfo = *srcinfo;


	return brow;
}
